<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">
<!--
  This file is generated by WebScrapBook and is not intended to be edited.
  Create frame.css and/or frame.js for customization.
-->
<html dir="{{ i18n('@@bidi_dir') }}" data-scrapbook-tree-page="frame">
<head>
<meta charset="UTF-8">
<title>{{ title }}</title>
<link rel="stylesheet" href="frame.css">
<script src="frame.js"></script>
</head>
<frameset cols="200,*">
<frame name="nav" src="map.html">
<frame name="view">
</frameset>
</html>
