<!DOCTYPE html>
<html>
<head>
<meta charset="UTF-8">
<meta name="viewport" content="width=device-width,initial-scale=1">
<title>{% block title %}{% endblock %}</title>
{%- block meta %}{% endblock %}
<link rel="stylesheet" type="text/css" href="{{ static_url('common.css') }}">
{%- block links %}{% endblock %}
{%- block scripts %}{% endblock %}
</head>
<body>{% block content %}{% endblock %}</body>
</html>