<!DOCTYPE html>
<html>
<head>
<meta charset="UTF-8">
<meta name="viewport" content="width=device-width,initial-scale=1">
<title>{{ title }}</title>
<style>
body {
  word-break: break-all;
}

.debug {
  color: gray;
}

.warn {
  color: #f0ad4e;
  font-weight: bold;
}

.error {
  color: red;
  font-weight: bold;
}

.critical {
  color: red;
  font-weight: bold;
}
</style>
</head>
<body>
{%- for msg in messages -%}
  {%- if msg.type != 'debug' or debug %}
    <span class="{{ msg.type }}">{{ msg.msg }}<br></span>
  {%- endif -%}
{%- endfor -%}
</body>
</html>
