{% extends "base.html" %}
{% block title %}Page {{ path }}{% endblock %}
{% block content %}
<header>
<h1 id="header" class="breadcrumbs">
{%- for label, subpath, sep, is_last in get_breadcrumbs(pathparts, base, sitename) -%}
  {%- if not is_last -%}
    <a href="{{ quote_path(subpath) }}">{{ label }}</a>{{ sep }}
  {%- else -%}
    <a>{{ label }}</a>
  {%- endif -%}
{%- endfor -%}
</h1>
</header>
<main>
{{ content|safe }}
</main>
{% endblock %}